'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var reduxActions = require('redux-actions');

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var capitalize = function capitalize(word) {
  return word.charAt(0).toUpperCase() + word.slice(1);
};

var validateInject = function validateInject(injected) {
  if (!Array.isArray(injected)) {
    throw Error('Inject should be an array of dependency names');
  }

  var hasNonStringInjects = injected.map(function (dep) {
    return typeof dep !== 'string';
  }).filter(Boolean).length > 0;

  if (hasNonStringInjects) {
    throw Error('Injected dependency names should be strings');
  }
};

var validateDuck = function validateDuck(_ref) {
  var name = _ref.name,
      state = _ref.state,
      inject = _ref.inject,
      actions = _ref.actions,
      reactions = _ref.reactions;

  if (!name) throw Error('Duck should have a name');

  if (state && !actions && !reactions) {
    throw Error('Duck with state should have reducers');
  }

  if (inject) validateInject(inject);
};

var createSelectors = function createSelectors(duck) {
  return Object.keys(duck.state).reduce(function (acc, key) {
    acc['get' + capitalize(key)] = function (state) {
      return state[duck.name][key];
    };
    return acc;
  }, {});
};

var handleThunks = function handleThunks(thunks, dependencies) {
  return Object.entries(thunks).reduce(function (acc, _ref2) {
    var actionName = _ref2[0],
        thunk = _ref2[1];

    acc[actionName] = function () {
      for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      return thunk.apply(undefined, args.concat([dependencies]));
    };
    return acc;
  }, {});
};

var createApiActionReducers = function createApiActionReducers(_ref3) {
  var _reducers;

  var types = _ref3.types,
      successField = _ref3.successField,
      _ref3$overrides = _ref3.overrides,
      overrides = _ref3$overrides === undefined ? {} : _ref3$overrides;

  var noOverrides = !overrides.loading && !overrides.failure && !overrides.success;

  if (!successField && noOverrides) {
    throw Error('You must provide the name of the field that is used for success payload');
  }

  var reducers = (_reducers = {}, _reducers[types.loading] = function (state) {
    return _extends({}, state, {
      error: null,
      hasError: false,
      isLoading: true
    });
  }, _reducers[types.success] = function (state, action) {
    var _extends2;

    return _extends({}, state, (_extends2 = {}, _extends2[successField] = action.payload, _extends2.error = null, _extends2.hasError = false, _extends2.isLoading = false, _extends2));
  }, _reducers[types.failure] = function (state, action) {
    return _extends({}, state, {
      error: action.payload,
      hasError: true,
      isLoading: false
    });
  }, _reducers);

  // Apply possible overrides
  if (overrides.loading) reducers[types.loading] = overrides.loading;
  if (overrides.success) reducers[types.success] = overrides.success;
  if (overrides.failure) reducers[types.failure] = overrides.failure;

  return reducers;
};

var createApiActionVariations = function createApiActionVariations(types) {
  var action = reduxActions.createAction(types.loading);
  action.success = reduxActions.createAction(types.success);
  action.fail = reduxActions.createAction(types.failure);
  return action;
};

function handleApiAction(arg, name, duckName) {
  var types = {
    loading: duckName + '/' + name,
    success: duckName + '/' + name + '/success',
    failure: duckName + '/' + name + '/failure'
  };

  var action = createApiActionVariations(types);

  // User can either provide only reducer field name for success case
  // or reducer overrides for `loading` / `success` / `failure` cases
  var reducers = typeof arg === 'string' ? createApiActionReducers({ types: types, successField: arg }) : createApiActionReducers({ types: types, overrides: arg });

  return { action: action, reducers: reducers, types: types };
}

var API_ACTION_IDENTIFIER = '__isApiAction__';

var isApiAction = function isApiAction(x) {
  return x && Object.prototype.hasOwnProperty.call(x, API_ACTION_IDENTIFIER);
};

var _extends$1 = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

// JSDoc typedefs
// TODO: maybe use TypeScript or Flow?

/**
 * @typedef {Object.<string, Function>} Action
 */

/**
 * @typedef {Object.<string, Function>} Selectors
 */

/**
 * @typedef {Object.<string, string>} Types
 */

/**
 * @typedef {Object} DuckDefinition
 * @property {String} name
 * @property {Object} state
 * @property {String[]} inject
 * @property {Function} actions
 * @property {Function} selectors
 * @property {Function} sagas
 */

/**
 * @typedef {Object} Duck
 * @property {Function} _run
 * @property {Function} _fillDeps
 * @property {Boolean} _created
 * @property {Function} getSagas
 * @property {Function} getReducer
 * @property {String} name
 * @property {Action} actions
 * @property {Types} types
 * @property {Selectors} selectors
 * @property {Object.<string, any>} initialState
 */

/**
 * Creates a ducks duck and returns chainable functions to define futher props.
 * @param {string} duckName
 * @param {string[]} typeList
 * @param {Object.<string, any>} duck
 * @returns {Duck}
 */
var createDuck = function createDuck(duck) {
  validateDuck(duck);

  var initialState = _extends$1({}, duck.state);
  var dependencies = {};
  var types = {};
  var reducerHandlers = {};
  var actions = {};
  var reducer = void 0;
  var sagas = [];

  // Mark injected deps to be filled later
  if (duck.inject) {
    duck.inject.forEach(function (depName) {
      dependencies[depName] = null;
    });
  }

  // Create types, actions and reducer handlers from `duck.actions`
  if (duck.actions) {
    // We need to provide `initialState` since it might be used in reducers
    Object.entries(duck.actions({ initialState: initialState })).forEach(function (_ref) {
      var actionName = _ref[0],
          reducerHandler = _ref[1];

      if (isApiAction(reducerHandler)) {
        // Handle async API actions
        var x = handleApiAction(reducerHandler.args, actionName, duck.name);
        types[actionName] = x.types;
        actions[actionName] = x.action;
        reducerHandlers = _extends$1({}, reducerHandlers, x.reducers);
      } else {
        // Register action type
        var actionType = duck.name + '/' + actionName;
        types[actionName] = actionType;

        // Create basic action
        actions[actionName] = reduxActions.createAction(actionType);
        reducerHandlers[actionType] = reducerHandler;
      }
    }, {});
  }

  // Handle thunks
  if (duck.thunks) {
    actions = _extends$1({}, actions, handleThunks(duck.thunks, dependencies));
  }

  // Auto-generate initial selectors for each state field
  var selectors = createSelectors(duck);

  // Add selectors defined by user
  if (duck.selectors) {
    var customSelectors = duck.selectors({ name: duck.name }) || {};
    selectors = _extends$1({}, selectors, customSelectors);
  }

  // Add simple `get` for selecting state fields by name
  var getSelector = function getSelector(key, state) {
    var statePart = state[duck.name];

    if (!Object.prototype.hasOwnProperty.call(statePart, key)) {
      throw Error('You tried to select a non-existent field \'' + key + '\' from state');
    }

    return statePart[key];
  };

  selectors = _extends$1({}, selectors, { get: getSelector });

  /**
   * Fills in the dependencies that were requested when calling inject.
   * @param {Object.<string, Duck>} deps
   */
  var _fillDeps = function _fillDeps(deps) {
    Object.keys(dependencies).forEach(function (dep) {
      if (deps[dep]) {
        dependencies[dep] = {
          // Don't spread all of dep but only what's needed
          types: deps[dep].types,
          actions: deps[dep].actions,
          selectors: deps[dep].selectors
        };
      } else {
        throw Error('There is no dependendy called \'' + dep + '\' for ' + duck.name + ' duck.' // eslint-disable-line
        );
      }
    });
  };

  /**
   * Run reducer and sagas functions with the necessary data.
   */
  var _run = function _run() {
    // Run duck reactions function to get rest of the reducer
    if (duck.reactions) {
      var fromDeps = duck.reactions({ initialState: initialState, deps: dependencies }) || {};
      reducerHandlers = _extends$1({}, reducerHandlers, fromDeps);
    }

    reducer = reduxActions.handleActions(reducerHandlers, initialState);

    if (duck.sagas) {
      sagas = duck.sagas({ types: types, deps: dependencies });
    }
  };

  /**
   * Get reducer for the duck.
   * @returns {Function}
   */
  var getReducer = function getReducer() {
    return reducer;
  };

  /**
   * Get sagas for the duck.
   * @returns {Array}
   */
  var getSagas = function getSagas() {
    return sagas || [];
  };

  return {
    name: duck.name,
    initialState: duck.state,
    types: types,
    actions: actions,
    selectors: selectors,
    getSagas: getSagas,
    getReducer: getReducer,
    _run: _run,
    _fillDeps: _fillDeps
  };
};

/**
 * Create ducks.
 * @param {Duck[]} ducks
 * @returns {Object.<string, Duck>}
 */
var initDucks = function initDucks() {
  var ducks = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];

  var ducksByName = ducks.reduce(function (acc, val) {
    acc[val.name] = val;
    return acc;
  }, {});

  // @ts-ignore
  Object.values(ducksByName).forEach(function (duck) {
    duck._fillDeps(ducksByName);
    duck._run();

    // Users don't need these props so delete them
    delete duck._fillDeps;
    delete duck._run;
  });

  var getAllReducers = function getAllReducers() {
    return Object.values(ducksByName).reduce(function (acc, val) {
      acc[val.name] = val.getReducer();
      return acc;
    }, {});
  };

  var getAllSagas = function getAllSagas() {
    return Object.values(ducksByName).reduce(function (acc, val) {
      return acc.concat.apply(acc, val.getSagas());
    }, []);
  };

  return _extends$1({}, ducksByName, {
    allReducers: getAllReducers(),
    allSagas: getAllSagas()
  });
};

var createApiAction = function createApiAction(args) {
  var _ref2;

  return _ref2 = {}, _ref2[API_ACTION_IDENTIFIER] = true, _ref2.args = args, _ref2;
};

exports.createDuck = createDuck;
exports.initDucks = initDucks;
exports.createApiAction = createApiAction;
